from .base_summary import Summary
from .binary_classification_summary import BinaryClassificationSummary
from .classification_summary import ClassificationSummary
from .regression_summary import RegressionSummary


import torch.nn as nn


class Reshape(nn.Module):
    def __init__(self, *args):
        super().__init__()
        self.shape = args

    def forward(self, x):
        return x.view(-1, *self.shape)



from .base_model import Model
from .sklean_model import SkModel
from .keras_model import KerasModel
# from .multi_model import
from .pytoch_model import PytorchModel
from .reinforcement_model import ReinforcementModel
from .auto_encoder_model import AutoEncoderModel



from typing import Tuple

import pandas as pd

from pandas_ml_utils.ml.data.splitting.splitter import Splitter


class NaiveSplitter(Splitter):

    def __init__(self, test_size=0.3):
        super().__init__()
        self.test_size = test_size

    def train_test_split(self, index: pd.Index) -> Tuple[pd.Index, pd.Index]:
        # we just split the sequence int past and recent data
        end_idx = int(len(index) * (1 - self.test_size))
        return index[0:end_idx], index[end_idx:]




from __future__ import annotations

import contextlib
import logging
import os
import tempfile
import uuid
from copy import deepcopy
from typing import List, Callable, TYPE_CHECKING, Tuple

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_common.utils import merge_kwargs, suitable_kwargs
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model

_log = logging.getLogger(__name__)


class AutoEncoderModel(Model):

    def __init__(self,
                 trainable_model: Model,
                 encoded_column_names: List[str],
                 encoder_provider: Callable[[Model], Callable[[np.ndarray], np.ndarray]],
                 decoder_provider: Callable[[Model], Callable[[np.ndarray], np.ndarray]],
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(trainable_model.features_and_labels, summary_provider, **kwargs)
        self._features_and_labels = trainable_model.features_and_labels
        self.trainable_model = trainable_model
        self.encoded_column_names = encoded_column_names
        self.encoder_provider = encoder_provider
        self.decoder_provider = decoder_provider
        self._predictor = None

    @property
    def features_and_labels(self):
        return self._features_and_labels

    def as_encoder(self) -> Model:
        fnl_copy = deepcopy(self.trainable_model.features_and_labels)
        fnl_copy.set_label_columns(self.encoded_column_names, True)

        return self._copy_keep_original_model(
            self.encoder_provider(self.trainable_model),
            fnl_copy
        )

    def as_decoder(self, decoder_features: List[Typing._Selector]) -> Model:
        fnl_copy = deepcopy(self.trainable_model.features_and_labels)
        fnl_copy._features = decoder_features

        return self._copy_keep_original_model(
            self.decoder_provider(self.trainable_model),
            fnl_copy
        )

    def fit_fold(self,
                 x: np.ndarray, y: np.ndarray,
                 x_val: np.ndarray, y_val: np.ndarray,
                 sample_weight_train: np.ndarray, sample_weight_test: np.ndarray,
                 **kwargs) -> float:
        loss = self.trainable_model.fit_fold(x, y,x_val, y_val, sample_weight_train, sample_weight_test, **kwargs)
        self._predictor = self.trainable_model.predict_sample
        return loss

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if x.ndim > 2 and x.shape[-1] == 1:
            x = x.reshape(x.shape[:-1])

        if self._predictor is None:
            return self._copy_keep_original_model(self.trainable_model.predict_sample,
                                                  self.trainable_model.features_and_labels)\
                       ._predictor(x, **kwargs)
        else:
            return self._predictor(x, **kwargs)

    def _copy_keep_original_model(self, new_predictior, new_features_and_labels):
        copy = AutoEncoderModel(
            self.trainable_model,
            self.encoded_column_names,
            self.encoder_provider,
            self.decoder_provider,
            self.summary_provider,
            **self.kwargs
        )

        copy._predictor = new_predictior
        copy._features_and_labels = new_features_and_labels

        return copy

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_features_and_labels']
        del state['_predictor']
        return state

    def __call__(self, *args, **kwargs):
        return AutoEncoderModel(
            self.trainable_model(*args, **kwargs),
            self.encoded_column_names,
            self.encoder_provider,
            self.decoder_provider,
            self.summary_provider,
            **self.kwargs
        )



import numpy as np
import pandas as pd
from unittest import TestCase

from pandas_ml_utils.ml.data.splitting import RandomSequences, DummySplitter
from pandas_ml_utils.ml.data.splitting.sampeling import DataGenerator, Sampler


class TestDataGenerator(TestCase):

    def test_generate_single_sample(self):
        data = pd.DataFrame({"a": np.arange(10)})
        test = (data, data, None)
        train = (data, data, None)
        gen = Sampler(train, test)

        samples = [s for s in gen.sample()]

        self.assertEqual(1, len(samples))

    def test_generate_multiple_sample(self):
        data = pd.DataFrame({"a": np.arange(10)})
        test = (data, data, None)
        train = (data, data, None)
        gen = Sampler(train, test, cross_validation=RandomSequences(0, 0.5, 12).cross_validation)

        samples = [s for s in gen.sample()]

        self.assertEqual(12, len(samples))

    def test_data_generator(self):
        pass

    def test_dummy_generator(self):
        features = pd.DataFrame({"a": np.arange(10)})
        targets = pd.DataFrame({"a": np.arange(10)})

        one = DataGenerator(DummySplitter(1), features, targets).complete_samples()
        two = DataGenerator(DummySplitter(2), features, targets).complete_samples()

        sampled1 = [s for s in one.sample()]
        sampled2 = [s for s in two.sample()]

        self.assertEqual(1, len(sampled1))
        self.assertEqual(2, len(sampled2))


"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.1.4'

import sys as _sys

from pandas_ml_common import *
from pandas_ml_utils.ml.data import *
from pandas_ml_utils.df.model import Model as DfModelExtension
from pandas_ml_utils.ml.model import *
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels

_log = logging.getLogger(__name__)
_log.debug(f"numpy version {np.__version__}")
_log.debug(f"pandas version {pd.__version__}")


setattr(PandasObject, "model", property(lambda self: DfModelExtension(self)))


if 'pandas_ml_quant' in _sys.modules:
    import pandas_ml_quant
    _log.warning(f"automatically imported pandas_ml_quant {pandas_ml_quant.__version__}")
